# This file is part of cloud-init. See LICENSE file for license information.

import copy
import os
import re
import shutil
from io import StringIO
from textwrap import dedent
from unittest import mock

import pytest
import yaml

from cloudinit import features, subp, util
from cloudinit.distros.parsers.sys_conf import SysConf
from cloudinit.net.activators import IfUpDownActivator
from tests.helpers import cloud_init_project_dir
from tests.unittests.helpers import dir2dict, get_distro

BASE_NET_CFG = """
auto lo
iface lo inet loopback

auto eth0
iface eth0 inet static
    address 192.168.1.5
    broadcast 192.168.1.0
    gateway 192.168.1.254
    netmask 255.255.255.0
    network 192.168.0.0

auto eth1
iface eth1 inet dhcp
"""

BASE_NET_CFG_FROM_V2 = """
auto lo
iface lo inet loopback

auto eth0
iface eth0 inet static
    address 192.168.1.5/24
    gateway 192.168.1.254

auto eth1
iface eth1 inet dhcp
"""

BASE_NET_CFG_IPV6 = """
auto lo
iface lo inet loopback

auto eth0
iface eth0 inet static
    address 192.168.1.5
    netmask 255.255.255.0
    network 192.168.0.0
    broadcast 192.168.1.0
    gateway 192.168.1.254

iface eth0 inet6 static
    address 2607:f0d0:1002:0011::2
    netmask 64
    gateway 2607:f0d0:1002:0011::1

iface eth1 inet static
    address 192.168.1.6
    netmask 255.255.255.0
    network 192.168.0.0
    broadcast 192.168.1.0
    gateway 192.168.1.254

iface eth1 inet6 static
    address 2607:f0d0:1002:0011::3
    netmask 64
    gateway 2607:f0d0:1002:0011::1
"""

V1_NET_CFG = {
    "config": [
        {
            "name": "eth0",
            "subnets": [
                {
                    "address": "192.168.1.5",
                    "broadcast": "192.168.1.0",
                    "gateway": "192.168.1.254",
                    "netmask": "255.255.255.0",
                    "type": "static",
                }
            ],
            "type": "physical",
        },
        {
            "name": "eth1",
            "subnets": [{"control": "auto", "type": "dhcp4"}],
            "type": "physical",
        },
    ],
    "version": 1,
}

V1_NET_CFG_WITH_DUPS = """\
# same value in interface specific dns and global dns
# should produce single entry in network file
version: 1
config:
    - type: physical
      name: eth0
      subnets:
          - type: static
            address: 192.168.0.102/24
            dns_nameservers: [1.2.3.4]
            dns_search: [test.com]
            interface: eth0
    - type: nameserver
      address: [1.2.3.4]
      search: [test.com]
"""

V1_NET_CFG_OUTPUT = """\
# This file is generated from information provided by the datasource.  Changes
# to it will not persist across an instance reboot.  To disable cloud-init's
# network configuration capabilities, write a file
# /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg with the following:
# network: {config: disabled}
auto lo
iface lo inet loopback

auto eth0
iface eth0 inet static
    address 192.168.1.5/24
    broadcast 192.168.1.0
    gateway 192.168.1.254

auto eth1
iface eth1 inet dhcp
"""

V1_NET_CFG_IPV6_OUTPUT = """\
# This file is generated from information provided by the datasource.  Changes
# to it will not persist across an instance reboot.  To disable cloud-init's
# network configuration capabilities, write a file
# /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg with the following:
# network: {config: disabled}
auto lo
iface lo inet loopback

auto eth0
iface eth0 inet6 static
    address 2607:f0d0:1002:0011::2/64
    gateway 2607:f0d0:1002:0011::1

auto eth1
iface eth1 inet dhcp
"""

V1_NET_CFG_IPV6 = {
    "config": [
        {
            "name": "eth0",
            "subnets": [
                {
                    "address": "2607:f0d0:1002:0011::2",
                    "gateway": "2607:f0d0:1002:0011::1",
                    "netmask": "64",
                    "type": "static6",
                }
            ],
            "type": "physical",
        },
        {
            "name": "eth1",
            "subnets": [{"control": "auto", "type": "dhcp4"}],
            "type": "physical",
        },
    ],
    "version": 1,
}


V1_TO_V2_NET_CFG_OUTPUT = """\
# This file is generated from information provided by the datasource.  Changes
# to it will not persist across an instance reboot.  To disable cloud-init's
# network configuration capabilities, write a file
# /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg with the following:
# network: {config: disabled}
network:
    version: 2
    ethernets:
        eth0:
            addresses:
            - 192.168.1.5/24
            routes:
            -   to: default
                via: 192.168.1.254
        eth1:
            dhcp4: true
"""

V1_TO_V2_NET_CFG_IPV6_OUTPUT = """\
# This file is generated from information provided by the datasource.  Changes
# to it will not persist across an instance reboot.  To disable cloud-init's
# network configuration capabilities, write a file
# /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg with the following:
# network: {config: disabled}
network:
    version: 2
    ethernets:
        eth0:
            addresses:
            - 2607:f0d0:1002:0011::2/64
            routes:
            -   to: default
                via: 2607:f0d0:1002:0011::1
        eth1:
            dhcp4: true
"""

V2_NET_CFG = {
    "ethernets": {
        "eth7": {"addresses": ["192.168.1.5/24"], "gateway4": "192.168.1.254"},
        "eth9": {"dhcp4": True},
    },
    "version": 2,
}


V2_TO_V2_NET_CFG_OUTPUT = """\
# This file is generated from information provided by the datasource.  Changes
# to it will not persist across an instance reboot.  To disable cloud-init's
# network configuration capabilities, write a file
# /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg with the following:
# network: {config: disabled}
network:
    ethernets:
        eth7:
            addresses:
            - 192.168.1.5/24
            gateway4: 192.168.1.254
        eth9:
            dhcp4: true
    version: 2
"""


V2_PASSTHROUGH_NET_CFG = {
    "ethernets": {
        "eth7": {
            "addresses": ["192.168.1.5/24"],
            "gateway4": "192.168.1.254",
            "routes": [{"to": "default", "via": "10.0.4.1", "metric": 100}],
        },
    },
    "version": 2,
}


V2_PASSTHROUGH_NET_CFG_OUTPUT = """\
# This file is generated from information provided by the datasource.  Changes
# to it will not persist across an instance reboot.  To disable cloud-init's
# network configuration capabilities, write a file
# /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg with the following:
# network: {config: disabled}
network:
    ethernets:
        eth7:
            addresses:
            - 192.168.1.5/24
            gateway4: 192.168.1.254
            routes:
            -   metric: 100
                to: default
                via: 10.0.4.1
    version: 2
"""


class WriteBuffer:
    def __init__(self):
        self.buffer = StringIO()
        self.mode = None
        self.omode = None

    def write(self, text):
        self.buffer.write(text)

    def __str__(self):
        return self.buffer.getvalue()


@pytest.fixture(autouse=True)
def system_is_snappy():
    mock.patch("cloudinit.util.system_is_snappy")


def assertCfgEquals(blob1, blob2):
    b1 = dict(SysConf(blob1.strip().splitlines()))
    b2 = dict(SysConf(blob2.strip().splitlines()))
    assert b1 == b2
    for k, v in b1.items():
        assert k in b2
    for k, v in b2.items():
        assert k in b1
    for k, v in b1.items():
        assert v == b2[k]


@pytest.fixture
def distro_freebsd(mocker):
    with open("tests/data/netinfo/freebsd-ifconfig-output", "r") as fh:
        ifs_txt = fh.read()
        mocker.patch(
            "cloudinit.distros.networking.subp.subp",
            return_value=(ifs_txt, None),
        )
        return get_distro("freebsd", renderers=["freebsd"])


@pytest.fixture(autouse=True)
def with_test_data(tmp_path, fake_filesystem_hook):
    shutil.copytree(
        str(cloud_init_project_dir("tests/data")),
        str(tmp_path / "tests/data"),
        dirs_exist_ok=True,
    )


@pytest.mark.usefixtures("fake_filesystem")
class TestNetCfgDistroFreeBSD:
    def _apply_and_verify_freebsd(
        self, apply_fn, config, tmp_path, expected_cfgs=None, bringup=False
    ):
        rootd = tmp_path
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.freebsd.available") as m_avail:
            m_avail.return_value = True
            util.ensure_dir(rootd / "etc")
            util.ensure_file(rootd / "etc/rc.conf")
            util.ensure_file(rootd / "etc/resolv.conf")
            apply_fn(config, bringup)

        results = dir2dict(
            str(rootd), filter=lambda fn: fn.startswith(str(rootd / "etc"))
        )
        for cfgpath, expected in expected_cfgs.items():
            print("----------")
            print(expected)
            print("^^^^ expected | rendered VVVVVVV")
            print(results[cfgpath])
            print("----------")
            assert set(expected.split("\n")) == set(
                results[cfgpath].split("\n")
            )
            assert 0o644 == get_mode(cfgpath, str(rootd))

    @mock.patch("cloudinit.net.get_interfaces_by_mac")
    def test_apply_network_config_freebsd_standard(
        self, ifaces_mac, distro_freebsd, tmp_path
    ):
        ifaces_mac.return_value = {
            "00:15:5d:4c:73:00": "eth0",
        }
        rc_conf_expected = """\
defaultrouter=192.168.1.254
ifconfig_eth0='inet 192.168.1.5 netmask 255.255.255.0'
ifconfig_eth1=DHCP
"""

        expected_cfgs = {
            "/etc/rc.conf": rc_conf_expected,
            "/etc/resolv.conf": "",
        }
        self._apply_and_verify_freebsd(
            distro_freebsd.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    @mock.patch("cloudinit.net.get_interfaces_by_mac")
    def test_apply_network_config_freebsd_ipv6_standard(
        self, ifaces_mac, distro_freebsd, tmp_path
    ):
        ifaces_mac.return_value = {
            "00:15:5d:4c:73:00": "eth0",
        }
        rc_conf_expected = """\
ipv6_defaultrouter=2607:f0d0:1002:0011::1
ifconfig_eth1=DHCP
ifconfig_eth0_ipv6='inet6 2607:f0d0:1002:0011::2/64'
"""

        expected_cfgs = {
            "/etc/rc.conf": rc_conf_expected,
            "/etc/resolv.conf": "",
        }
        self._apply_and_verify_freebsd(
            distro_freebsd.apply_network_config,
            V1_NET_CFG_IPV6,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    @mock.patch("cloudinit.net.get_interfaces_by_mac")
    def test_apply_network_config_freebsd_ifrename(
        self, ifaces_mac, distro_freebsd, tmp_path
    ):
        ifaces_mac.return_value = {
            "00:15:5d:4c:73:00": "vtnet0",
        }
        rc_conf_expected = """\
ifconfig_vtnet0_name=eth0
defaultrouter=192.168.1.254
ifconfig_eth0='inet 192.168.1.5 netmask 255.255.255.0'
ifconfig_eth1=DHCP
"""

        V1_NET_CFG_RENAME = copy.deepcopy(V1_NET_CFG)
        V1_NET_CFG_RENAME["config"][0]["mac_address"] = "00:15:5d:4c:73:00"

        expected_cfgs = {
            "/etc/rc.conf": rc_conf_expected,
            "/etc/resolv.conf": "",
        }
        self._apply_and_verify_freebsd(
            distro_freebsd.apply_network_config,
            V1_NET_CFG_RENAME,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    @mock.patch("cloudinit.net.get_interfaces_by_mac")
    def test_apply_network_config_freebsd_nameserver(
        self, ifaces_mac, distro_freebsd, tmp_path
    ):
        ifaces_mac.return_value = {
            "00:15:5d:4c:73:00": "eth0",
        }

        V1_NET_CFG_DNS = copy.deepcopy(V1_NET_CFG)
        ns = ["1.2.3.4"]
        V1_NET_CFG_DNS["config"][0]["subnets"][0]["dns_nameservers"] = ns
        expected_cfgs = {"/etc/resolv.conf": "nameserver 1.2.3.4\n"}
        self._apply_and_verify_freebsd(
            distro_freebsd.apply_network_config,
            V1_NET_CFG_DNS,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )


@pytest.fixture
def distro_eni():
    return get_distro("ubuntu", renderers=["eni"], activators=["eni"])


@pytest.fixture
def m_activators_subp(mocker):
    mocker.patch("cloudinit.net.activators.subp.subp", return_value=("", ""))


@pytest.mark.usefixtures("fake_filesystem", "m_activators_subp")
class TestNetCfgDistroUbuntuEni:
    def eni_path(self):
        return "/etc/network/interfaces.d/50-cloud-init.cfg"

    def rules_path(self):
        return "/etc/udev/rules.d/70-persistent-net.rules"

    def _apply_and_verify_eni(
        self,
        apply_fn,
        config,
        tmp_path,
        expected_cfgs=None,
        bringup=False,
        previous_files=(),
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.eni.available") as m_avail:
            m_avail.return_value = True
            path_modes = {}
            for previous_path, content, mode in previous_files:
                util.write_file(previous_path, content, mode=mode)
                path_modes[previous_path] = mode
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected in expected_cfgs.items():
            print("----------")
            print(expected)
            print("^^^^ expected | rendered VVVVVVV")
            print(results[cfgpath])
            print("----------")
            assert expected == results[cfgpath]
            assert path_modes.get(cfgpath, 0o644) == get_mode(
                cfgpath, str(tmp_path)
            )

    def test_apply_network_config_and_bringup_filters_priority_eni_ub(
        self, distro_eni, tmp_path
    ):
        """Network activator search priority can be overridden from config."""
        expected_cfgs = {
            self.eni_path(): V1_NET_CFG_OUTPUT,
        }

        with mock.patch(
            "cloudinit.net.activators.select_activator"
        ) as select_activator:
            select_activator.return_value = IfUpDownActivator
            self._apply_and_verify_eni(
                distro_eni.apply_network_config,
                V1_NET_CFG,
                tmp_path,
                expected_cfgs=expected_cfgs.copy(),
                bringup=True,
            )
            # 2nd call to select_activator via distro.network_activator prop
            assert IfUpDownActivator == distro_eni.network_activator
        assert [
            mock.call(priority=["eni"])
        ] * 2 == select_activator.call_args_list

    def test_apply_network_config_and_bringup_activator_defaults_ub(
        self, tmp_path
    ):
        """Network activator search priority defaults when unspecified."""
        expected_cfgs = {
            self.eni_path(): V1_NET_CFG_OUTPUT,
        }
        # Don't set activators to see DEFAULT_PRIORITY
        distro = get_distro("ubuntu", renderers=["eni"])
        with mock.patch(
            "cloudinit.net.activators.select_activator"
        ) as select_activator:
            select_activator.return_value = IfUpDownActivator
            self._apply_and_verify_eni(
                distro.apply_network_config,
                V1_NET_CFG,
                tmp_path,
                expected_cfgs=expected_cfgs.copy(),
                bringup=True,
            )
            # 2nd call to select_activator via distro.network_activator prop
            assert IfUpDownActivator == distro.network_activator
        assert [
            mock.call(priority=None)
        ] * 2 == select_activator.call_args_list

    def test_apply_network_config_eni_ub(self, distro_eni, tmp_path):
        expected_cfgs = {
            self.eni_path(): V1_NET_CFG_OUTPUT,
            self.rules_path(): "",
        }
        self._apply_and_verify_eni(
            distro_eni.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
            previous_files=((self.rules_path(), "something", 0o660),),
        )

    def test_apply_network_config_ipv6_ub(self, distro_eni, tmp_path):
        expected_cfgs = {self.eni_path(): V1_NET_CFG_IPV6_OUTPUT}
        self._apply_and_verify_eni(
            distro_eni.apply_network_config,
            V1_NET_CFG_IPV6,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )


@pytest.fixture
def distro_netplan():
    return get_distro("ubuntu", renderers=["netplan"])


@pytest.fixture
def m_netplan_subp(mocker):
    mocker.patch("cloudinit.net.netplan.subp.subp", return_value=("", ""))


@pytest.mark.usefixtures("fake_filesystem", "m_netplan_subp")
class TestNetCfgDistroUbuntuNetplan:
    DEV_LIST = ["eth0", "lo"]

    def _apply_and_verify_netplan(
        self,
        apply_fn,
        config,
        tmp_path,
        expected_cfgs=None,
        bringup=False,
        previous_files=(),
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.netplan.available", return_value=True):
            with mock.patch(
                "cloudinit.net.netplan.get_devicelist",
                return_value=self.DEV_LIST,
            ):
                for previous_path, content, mode in previous_files:
                    util.write_file(previous_path, content, mode=mode)
                apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected, mode in expected_cfgs:
            print("----------")
            print(expected)
            print("^^^^ expected | rendered VVVVVVV")
            print(results[cfgpath])
            print("----------")
            assert expected == results[cfgpath]
            assert mode == get_mode(cfgpath, str(tmp_path))

    def netplan_path(self):
        return "/etc/netplan/50-cloud-init.yaml"

    def test_apply_network_config_v1_to_netplan_ub(
        self, distro_netplan, tmp_path
    ):
        expected_cfgs = (
            (self.netplan_path(), V1_TO_V2_NET_CFG_OUTPUT, 0o600),
        )

        self._apply_and_verify_netplan(
            distro_netplan.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs,
        )

    def test_apply_network_config_v1_ipv6_to_netplan_ub(
        self, distro_netplan, tmp_path
    ):
        expected_cfgs = (
            (self.netplan_path(), V1_TO_V2_NET_CFG_IPV6_OUTPUT, 0o600),
        )

        self._apply_and_verify_netplan(
            distro_netplan.apply_network_config,
            V1_NET_CFG_IPV6,
            tmp_path,
            expected_cfgs=expected_cfgs,
        )

    def test_apply_network_config_v2_passthrough_ub(
        self, distro_netplan, tmp_path
    ):
        expected_cfgs = (
            (self.netplan_path(), V2_TO_V2_NET_CFG_OUTPUT, 0o600),
        )
        self._apply_and_verify_netplan(
            distro_netplan.apply_network_config,
            V2_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs,
        )

    def test_apply_network_config_v2_passthrough_retain_orig_perms(
        self, distro_netplan, tmp_path
    ):
        """Custom permissions on existing netplan is kept when more strict."""
        expected_cfgs = (
            (self.netplan_path(), V2_TO_V2_NET_CFG_OUTPUT, 0o640),
        )
        with mock.patch.object(
            features, "NETPLAN_CONFIG_ROOT_READ_ONLY", False
        ):
            # When NETPLAN_CONFIG_ROOT_READ_ONLY is False default perms are 644
            # we keep 640 because it's more strict.
            # 1640 is used to assert sticky bit preserved across write
            self._apply_and_verify_netplan(
                distro_netplan.apply_network_config,
                V2_NET_CFG,
                tmp_path,
                expected_cfgs=expected_cfgs,
                previous_files=(
                    ("/etc/netplan/50-cloud-init.yaml", "a", 0o640),
                ),
            )

    def test_apply_network_config_v2_passthrough_ub_old_behavior(
        self, distro_netplan, tmp_path
    ):
        """Kinetic and earlier have 50-cloud-init.yaml world-readable"""
        expected_cfgs = (
            (self.netplan_path(), V2_TO_V2_NET_CFG_OUTPUT, 0o644),
        )
        with mock.patch.object(
            features, "NETPLAN_CONFIG_ROOT_READ_ONLY", False
        ):
            self._apply_and_verify_netplan(
                distro_netplan.apply_network_config,
                V2_NET_CFG,
                tmp_path,
                expected_cfgs=expected_cfgs,
            )

    def test_apply_network_config_v2_full_passthrough_ub(
        self, caplog, distro_netplan, tmp_path
    ):
        expected_cfgs = (
            (self.netplan_path(), V2_PASSTHROUGH_NET_CFG_OUTPUT, 0o600),
        )
        self._apply_and_verify_netplan(
            distro_netplan.apply_network_config,
            V2_PASSTHROUGH_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs,
        )
        assert "Passthrough netplan v2 config" in caplog.text
        assert (
            "Selected renderer 'netplan' from priority list: ['netplan']"
            in caplog.text
        )


@pytest.fixture
def distro_redhat():
    return get_distro("rhel", renderers=["sysconfig"])


@pytest.mark.usefixtures("fake_filesystem")
class TestNetCfgDistroRedhat:

    def ifcfg_path(self, ifname):
        return "/etc/sysconfig/network-scripts/ifcfg-%s" % ifname

    def control_path(self):
        return "/etc/sysconfig/network"

    def _apply_and_verify(
        self,
        apply_fn,
        config,
        tmp_path,
        expected_cfgs=None,
        bringup=False,
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.sysconfig.available") as m_avail:
            m_avail.return_value = True
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected in expected_cfgs.items():
            assertCfgEquals(expected, results[cfgpath])
            assert 0o644 == get_mode(cfgpath, str(tmp_path))

    def test_apply_network_config_rh(self, distro_redhat, tmp_path):
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=none
                DEFROUTE=yes
                DEVICE=eth0
                GATEWAY=192.168.1.254
                IPADDR=192.168.1.5
                NETMASK=255.255.255.0
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.ifcfg_path("eth1"): dedent(
                """\
                BOOTPROTO=dhcp
                DEVICE=eth1
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.control_path(): dedent(
                """\
                NETWORKING=yes
                """
            ),
        }
        # rh_distro.apply_network_config(V1_NET_CFG, False)
        self._apply_and_verify(
            distro_redhat.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    def test_apply_network_config_ipv6_rh(self, distro_redhat, tmp_path):
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=none
                DEFROUTE=yes
                DEVICE=eth0
                IPV6ADDR=2607:f0d0:1002:0011::2/64
                IPV6INIT=yes
                IPV6_AUTOCONF=no
                IPV6_DEFAULTGW=2607:f0d0:1002:0011::1
                IPV6_FORCE_ACCEPT_RA=no
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.ifcfg_path("eth1"): dedent(
                """\
                BOOTPROTO=dhcp
                DEVICE=eth1
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.control_path(): dedent(
                """\
                NETWORKING=yes
                NETWORKING_IPV6=yes
                IPV6_AUTOCONF=no
                """
            ),
        }
        # rh_distro.apply_network_config(V1_NET_CFG_IPV6, False)
        self._apply_and_verify(
            distro_redhat.apply_network_config,
            V1_NET_CFG_IPV6,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    def test_sysconfig_network_no_overwite_ipv6_rh(
        self, distro_redhat, tmp_path
    ):
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=none
                DEFROUTE=yes
                DEVICE=eth0
                IPV6ADDR=2607:f0d0:1002:0011::2/64
                IPV6INIT=yes
                IPV6_AUTOCONF=no
                IPV6_DEFAULTGW=2607:f0d0:1002:0011::1
                IPV6_FORCE_ACCEPT_RA=no
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.ifcfg_path("eth1"): dedent(
                """\
                BOOTPROTO=dhcp
                DEVICE=eth1
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.control_path(): dedent(
                """\
                NETWORKING=yes
                NETWORKING_IPV6=yes
                IPV6_AUTOCONF=no
                NOZEROCONF=yes
                """
            ),
        }
        file_mode = 0o644
        # pre-existing config in /etc/sysconfig/network should not be removed
        util.write_file(
            self.control_path(),
            "".join("NOZEROCONF=yes") + "\n",
            file_mode,
        )

        self._apply_and_verify(
            distro_redhat.apply_network_config,
            V1_NET_CFG_IPV6,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    def test_vlan_render_unsupported(self, distro_redhat, tmp_path):
        """Render officially unsupported vlan names."""
        cfg = {
            "version": 2,
            "ethernets": {
                "eth0": {
                    "addresses": ["192.10.1.2/24"],
                    "match": {"macaddress": "00:16:3e:60:7c:df"},
                }
            },
            "vlans": {
                "infra0": {
                    "addresses": ["10.0.1.2/16"],
                    "id": 1001,
                    "link": "eth0",
                }
            },
        }
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=none
                DEVICE=eth0
                HWADDR=00:16:3e:60:7c:df
                IPADDR=192.10.1.2
                NETMASK=255.255.255.0
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.ifcfg_path("infra0"): dedent(
                """\
                BOOTPROTO=none
                DEVICE=infra0
                IPADDR=10.0.1.2
                NETMASK=255.255.0.0
                ONBOOT=yes
                PHYSDEV=eth0
                USERCTL=no
                VLAN=yes
                """
            ),
            self.control_path(): dedent(
                """\
                NETWORKING=yes
                """
            ),
        }
        self._apply_and_verify(
            distro_redhat.apply_network_config,
            cfg,
            tmp_path,
            expected_cfgs=expected_cfgs,
        )

    def test_vlan_render(self, distro_redhat, tmp_path):
        cfg = {
            "version": 2,
            "ethernets": {"eth0": {"addresses": ["192.10.1.2/24"]}},
            "vlans": {
                "eth0.1001": {
                    "addresses": ["10.0.1.2/16"],
                    "id": 1001,
                    "link": "eth0",
                }
            },
        }
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=none
                DEVICE=eth0
                IPADDR=192.10.1.2
                NETMASK=255.255.255.0
                ONBOOT=yes
                TYPE=Ethernet
                USERCTL=no
                """
            ),
            self.ifcfg_path("eth0.1001"): dedent(
                """\
                BOOTPROTO=none
                DEVICE=eth0.1001
                IPADDR=10.0.1.2
                NETMASK=255.255.0.0
                ONBOOT=yes
                PHYSDEV=eth0
                USERCTL=no
                VLAN=yes
                """
            ),
            self.control_path(): dedent(
                """\
                NETWORKING=yes
                """
            ),
        }
        self._apply_and_verify(
            distro_redhat.apply_network_config,
            cfg,
            tmp_path,
            expected_cfgs=expected_cfgs,
        )


@pytest.fixture
def distro_opensuse():
    return get_distro("opensuse", renderers=["sysconfig"])


@pytest.mark.usefixtures("fake_filesystem")
class TestNetCfgDistroOpensuse:
    def ifcfg_path(self, ifname):
        return "/etc/sysconfig/network/ifcfg-%s" % ifname

    def _apply_and_verify(
        self, apply_fn, config, tmp_path, expected_cfgs=None, bringup=False
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.sysconfig.available") as m_avail:
            m_avail.return_value = True
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected in expected_cfgs.items():
            assertCfgEquals(expected, results[cfgpath])
            assert 0o644 == get_mode(cfgpath, str(tmp_path))

    def test_apply_network_config_opensuse(self, distro_opensuse, tmp_path):
        """Opensuse uses apply_network_config and renders sysconfig"""
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=static
                IPADDR=192.168.1.5
                NETMASK=255.255.255.0
                STARTMODE=auto
                """
            ),
            self.ifcfg_path("eth1"): dedent(
                """\
                BOOTPROTO=dhcp4
                STARTMODE=auto
                """
            ),
        }
        self._apply_and_verify(
            distro_opensuse.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )

    def test_apply_network_config_ipv6_opensuse(
        self, distro_opensuse, tmp_path
    ):
        """Opensuse uses apply_network_config and renders sysconfig w/ipv6"""
        expected_cfgs = {
            self.ifcfg_path("eth0"): dedent(
                """\
                BOOTPROTO=static
                IPADDR6=2607:f0d0:1002:0011::2/64
                STARTMODE=auto
            """
            ),
            self.ifcfg_path("eth1"): dedent(
                """\
                BOOTPROTO=dhcp4
                STARTMODE=auto
            """
            ),
        }
        self._apply_and_verify(
            distro_opensuse.apply_network_config,
            V1_NET_CFG_IPV6,
            tmp_path,
            expected_cfgs=expected_cfgs.copy(),
        )


@pytest.fixture
def distro_arch():
    return get_distro("arch", renderers=["netplan"])


@pytest.mark.usefixtures("fake_filesystem", "m_netplan_subp")
class TestNetCfgDistroArch:
    def _apply_and_verify(
        self,
        apply_fn,
        config,
        tmp_path,
        expected_cfgs=None,
        bringup=False,
        with_netplan=False,
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch(
            "cloudinit.net.netplan.available", return_value=with_netplan
        ):
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        mode = 0o600 if with_netplan else 0o644
        for cfgpath, expected in expected_cfgs.items():
            print("----------")
            print(expected)
            print("^^^^ expected | rendered VVVVVVV")
            print(results[cfgpath])
            print("----------")
            assert expected == results[cfgpath]
            assert mode == get_mode(cfgpath, str(tmp_path))

    def netctl_path(self, iface):
        return "/etc/netctl/%s" % iface

    def netplan_path(self):
        return "/etc/netplan/50-cloud-init.yaml"

    def test_apply_network_config_v1_with_netplan(self, distro_arch, tmp_path):
        expected_cfgs = {
            self.netplan_path(): dedent(
                """\
                # generated by cloud-init
                network:
                    version: 2
                    ethernets:
                        eth0:
                            addresses:
                            - 192.168.1.5/24
                            routes:
                            -   to: default
                                via: 192.168.1.254
                        eth1:
                            dhcp4: true
                """
            ),
        }

        with mock.patch(
            "cloudinit.net.netplan.get_devicelist", return_value=[]
        ):
            self._apply_and_verify(
                distro_arch.apply_network_config,
                V1_NET_CFG,
                tmp_path,
                expected_cfgs=expected_cfgs.copy(),
                with_netplan=True,
            )


@pytest.fixture
def distro_photon(mocker):
    mocker.patch("cloudinit.net.networkd.util.chownbyname")
    return get_distro("photon", renderers=["networkd"])


@pytest.mark.usefixtures("fake_filesystem")
class TestNetCfgDistroPhoton:
    def create_conf_dict(self, contents):
        content_dict = {}
        for line in contents:
            if line:
                line = line.strip()
                if line and re.search(r"^\[(.+)\]$", line):
                    content_dict[line] = []
                    key = line
                elif line:
                    assert key
                    content_dict[key].append(line)

        return content_dict

    def compare_dicts(self, actual, expected):
        for k, v in actual.items():
            assert sorted(expected[k]) == sorted(v)

    def _apply_and_verify(
        self, apply_fn, config, tmp_path, expected_cfgs=None, bringup=False
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.networkd.available") as m_avail:
            m_avail.return_value = True
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected in expected_cfgs.items():
            actual = self.create_conf_dict(results[cfgpath].splitlines())
            self.compare_dicts(actual, expected)
            assert 0o644 == get_mode(cfgpath, str(tmp_path))

    def nwk_file_path(self, ifname):
        return "/etc/systemd/network/10-cloud-init-%s.network" % ifname

    def net_cfg_1(self, ifname):
        ret = (
            """\
        [Match]
        Name=%s
        [Network]
        DHCP=no
        [Address]
        Address=192.168.1.5/24
        [Route]
        Gateway=192.168.1.254"""
            % ifname
        )
        return ret

    def net_cfg_2(self, ifname):
        ret = (
            """\
        [Match]
        Name=%s
        [Network]
        DHCP=ipv4"""
            % ifname
        )
        return ret

    def test_photon_network_config_v1(self, distro_photon, tmp_path):
        tmp = self.net_cfg_1("eth0").splitlines()
        expected_eth0 = self.create_conf_dict(tmp)

        tmp = self.net_cfg_2("eth1").splitlines()
        expected_eth1 = self.create_conf_dict(tmp)

        expected_cfgs = {
            self.nwk_file_path("eth0"): expected_eth0,
            self.nwk_file_path("eth1"): expected_eth1,
        }

        self._apply_and_verify(
            distro_photon.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs.copy(),
        )

    def test_photon_network_config_v2(self, distro_photon, tmp_path):
        tmp = self.net_cfg_1("eth7").splitlines()
        expected_eth7 = self.create_conf_dict(tmp)

        tmp = self.net_cfg_2("eth9").splitlines()
        expected_eth9 = self.create_conf_dict(tmp)

        expected_cfgs = {
            self.nwk_file_path("eth7"): expected_eth7,
            self.nwk_file_path("eth9"): expected_eth9,
        }

        self._apply_and_verify(
            distro_photon.apply_network_config,
            V2_NET_CFG,
            tmp_path,
            expected_cfgs.copy(),
        )

    def test_photon_network_config_v1_with_duplicates(
        self, distro_photon, tmp_path
    ):
        expected = """\
        [Match]
        Name=eth0
        [Network]
        DHCP=no
        DNS=1.2.3.4
        Domains=test.com
        [Address]
        Address=192.168.0.102/24"""

        net_cfg = yaml.safe_load(V1_NET_CFG_WITH_DUPS)

        expected = self.create_conf_dict(expected.splitlines())
        expected_cfgs = {
            self.nwk_file_path("eth0"): expected,
        }

        self._apply_and_verify(
            distro_photon.apply_network_config,
            net_cfg,
            tmp_path,
            expected_cfgs.copy(),
        )


@pytest.fixture
def distro_mariner(mocker):
    mocker.patch("cloudinit.net.networkd.util.chownbyname")
    return get_distro("mariner", renderers=["networkd"])


@pytest.mark.usefixtures("fake_filesystem")
class TestNetCfgDistroMariner:
    def create_conf_dict(self, contents):
        content_dict = {}
        for line in contents:
            if line:
                line = line.strip()
                if line and re.search(r"^\[(.+)\]$", line):
                    content_dict[line] = []
                    key = line
                elif line:
                    assert key
                    content_dict[key].append(line)

        return content_dict

    def compare_dicts(self, actual, expected):
        for k, v in actual.items():
            assert sorted(expected[k]) == sorted(v)

    def _apply_and_verify(
        self, apply_fn, config, tmp_path, expected_cfgs=None, bringup=False
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.networkd.available") as m_avail:
            m_avail.return_value = True
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected in expected_cfgs.items():
            actual = self.create_conf_dict(results[cfgpath].splitlines())
            self.compare_dicts(actual, expected)
            assert 0o644 == get_mode(cfgpath, str(tmp_path))

    def nwk_file_path(self, ifname):
        return "/etc/systemd/network/10-cloud-init-%s.network" % ifname

    def net_cfg_1(self, ifname):
        ret = (
            """\
        [Match]
        Name=%s
        [Network]
        DHCP=no
        [Address]
        Address=192.168.1.5/24
        [Route]
        Gateway=192.168.1.254"""
            % ifname
        )
        return ret

    def net_cfg_2(self, ifname):
        ret = (
            """\
        [Match]
        Name=%s
        [Network]
        DHCP=ipv4"""
            % ifname
        )
        return ret

    def test_mariner_network_config_v1(self, distro_mariner, tmp_path):
        tmp = self.net_cfg_1("eth0").splitlines()
        expected_eth0 = self.create_conf_dict(tmp)

        tmp = self.net_cfg_2("eth1").splitlines()
        expected_eth1 = self.create_conf_dict(tmp)

        expected_cfgs = {
            self.nwk_file_path("eth0"): expected_eth0,
            self.nwk_file_path("eth1"): expected_eth1,
        }

        self._apply_and_verify(
            distro_mariner.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs.copy(),
        )

    def test_mariner_network_config_v2(self, distro_mariner, tmp_path):
        tmp = self.net_cfg_1("eth7").splitlines()
        expected_eth7 = self.create_conf_dict(tmp)

        tmp = self.net_cfg_2("eth9").splitlines()
        expected_eth9 = self.create_conf_dict(tmp)

        expected_cfgs = {
            self.nwk_file_path("eth7"): expected_eth7,
            self.nwk_file_path("eth9"): expected_eth9,
        }

        self._apply_and_verify(
            distro_mariner.apply_network_config,
            V2_NET_CFG,
            tmp_path,
            expected_cfgs.copy(),
        )

    def test_mariner_network_config_v1_with_duplicates(
        self, distro_mariner, tmp_path
    ):
        expected = """\
        [Match]
        Name=eth0
        [Network]
        DHCP=no
        DNS=1.2.3.4
        Domains=test.com
        [Address]
        Address=192.168.0.102/24"""

        net_cfg = yaml.safe_load(V1_NET_CFG_WITH_DUPS)

        expected = self.create_conf_dict(expected.splitlines())
        expected_cfgs = {
            self.nwk_file_path("eth0"): expected,
        }

        self._apply_and_verify(
            distro_mariner.apply_network_config,
            net_cfg,
            tmp_path,
            expected_cfgs.copy(),
        )


@pytest.fixture
def distro_azurelinux(mocker):
    mocker.patch("cloudinit.net.networkd.util.chownbyname")
    return get_distro("azurelinux", renderers=["networkd"])


@pytest.mark.usefixtures("fake_filesystem")
class TestNetCfgDistroAzureLinux:
    def create_conf_dict(self, contents):
        content_dict = {}
        for line in contents:
            if line:
                line = line.strip()
                if line and re.search(r"^\[(.+)\]$", line):
                    content_dict[line] = []
                    key = line
                elif line:
                    assert key
                    content_dict[key].append(line)

        return content_dict

    def compare_dicts(self, actual, expected):
        for k, v in actual.items():
            assert sorted(expected[k]) == sorted(v)

    def _apply_and_verify(
        self, apply_fn, config, tmp_path, expected_cfgs=None, bringup=False
    ):
        if not expected_cfgs:
            raise ValueError("expected_cfg must not be None")

        with mock.patch("cloudinit.net.networkd.available") as m_avail:
            m_avail.return_value = True
            apply_fn(config, bringup)

        results = dir2dict(
            str(tmp_path),
            filter=lambda fn: fn.startswith(str(tmp_path / "etc")),
        )
        for cfgpath, expected in expected_cfgs.items():
            actual = self.create_conf_dict(results[cfgpath].splitlines())
            self.compare_dicts(actual, expected)
            assert 0o644 == get_mode(cfgpath, str(tmp_path))

    def nwk_file_path(self, ifname):
        return "/etc/systemd/network/10-cloud-init-%s.network" % ifname

    def net_cfg_1(self, ifname):
        ret = (
            """\
        [Match]
        Name=%s
        [Network]
        DHCP=no
        [Address]
        Address=192.168.1.5/24
        [Route]
        Gateway=192.168.1.254"""
            % ifname
        )
        return ret

    def net_cfg_2(self, ifname):
        ret = (
            """\
        [Match]
        Name=%s
        [Network]
        DHCP=ipv4"""
            % ifname
        )
        return ret

    def test_azurelinux_network_config_v1(self, distro_azurelinux, tmp_path):
        tmp = self.net_cfg_1("eth0").splitlines()
        expected_eth0 = self.create_conf_dict(tmp)

        tmp = self.net_cfg_2("eth1").splitlines()
        expected_eth1 = self.create_conf_dict(tmp)

        expected_cfgs = {
            self.nwk_file_path("eth0"): expected_eth0,
            self.nwk_file_path("eth1"): expected_eth1,
        }

        self._apply_and_verify(
            distro_azurelinux.apply_network_config,
            V1_NET_CFG,
            tmp_path,
            expected_cfgs.copy(),
        )

    def test_azurelinux_network_config_v2(self, distro_azurelinux, tmp_path):
        tmp = self.net_cfg_1("eth7").splitlines()
        expected_eth7 = self.create_conf_dict(tmp)

        tmp = self.net_cfg_2("eth9").splitlines()
        expected_eth9 = self.create_conf_dict(tmp)

        expected_cfgs = {
            self.nwk_file_path("eth7"): expected_eth7,
            self.nwk_file_path("eth9"): expected_eth9,
        }

        self._apply_and_verify(
            distro_azurelinux.apply_network_config,
            V2_NET_CFG,
            tmp_path,
            expected_cfgs.copy(),
        )

    def test_azurelinux_network_config_v1_with_duplicates(
        self, distro_azurelinux, tmp_path
    ):
        expected = """\
        [Match]
        Name=eth0
        [Network]
        DHCP=no
        DNS=1.2.3.4
        Domains=test.com
        [Address]
        Address=192.168.0.102/24"""

        net_cfg = yaml.safe_load(V1_NET_CFG_WITH_DUPS)

        expected = self.create_conf_dict(expected.splitlines())
        expected_cfgs = {
            self.nwk_file_path("eth0"): expected,
        }

        self._apply_and_verify(
            distro_azurelinux.apply_network_config,
            net_cfg,
            tmp_path,
            expected_cfgs.copy(),
        )


def get_mode(path, target=None):
    # Mask upper st_mode bits like S_IFREG bit preserve sticky and isuid/osgid
    return os.stat(subp.target_path(target, path)).st_mode & 0o777



# This file is part of cloud-init. See LICENSE file for license information.

from tests.unittests.helpers import get_distro


class TestSLES:
    def test_get_distro(self):
        distro = get_distro("sles")
        assert distro.osfamily == "suse"



# This file is part of cloud-init. See LICENSE file for license information.

import copy
import os

import pytest

from cloudinit import atomic_helper, safeyaml, stages, util
from cloudinit.config.modules import Modules
from cloudinit.settings import PER_INSTANCE
from cloudinit.sources import NetworkConfigSource
from tests.unittests.helpers import replicate_test_root


@pytest.fixture(autouse=True)
def replicate_root(tmp_path):
    replicate_test_root("simple_ubuntu", str(tmp_path))


@pytest.fixture(autouse=True)
def cfg(mocker, tmp_path):
    # root group doesn't exist everywhere
    mocker.patch("cloudinit.util.chownbyname")

    new_root = str(tmp_path)
    _cfg = {
        "datasource_list": ["None"],
        "runcmd": ["ls /etc"],  # test ALL_DISTROS
        "spacewalk": {},  # test non-ubuntu distros module definition
        "system_info": {
            "paths": {"run_dir": new_root},
            "distro": "ubuntu",
        },
        "write_files": [
            {
                "path": "/etc/blah.ini",
                "content": "blah",
                "permissions": 0o755,
            },
        ],
        "cloud_init_modules": ["write_files", "spacewalk", "runcmd"],
    }
    cloud_cfg = safeyaml.dumps(_cfg)
    util.ensure_dir(os.path.join(new_root, "etc", "cloud"))
    util.write_file(
        os.path.join(new_root, "etc", "cloud", "cloud.cfg"), cloud_cfg
    )
    return _cfg


@pytest.mark.usefixtures("fake_filesystem")
class TestSimpleRun:
    def test_none_ds_populates_var_lib_cloud(self):
        """Init and run_section default behavior creates appropriate dirs."""
        # Now start verifying whats created
        netcfg = {
            "version": 1,
            "config": [{"type": "physical", "name": "eth9"}],
        }

        def fake_network_config():
            return netcfg, NetworkConfigSource.FALLBACK

        assert not os.path.exists("/var/lib/cloud")
        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        assert os.path.exists("/var/lib/cloud")
        for d in ["scripts", "seed", "instances", "handlers", "sem", "data"]:
            assert os.path.isdir(os.path.join("/var/lib/cloud", d))

        initer.fetch()
        assert not os.path.islink("var/lib/cloud/instance")
        iid = initer.instancify()
        assert iid == "iid-datasource-none"
        initer.update()
        assert os.path.islink("var/lib/cloud/instance")
        initer._find_networking_config = fake_network_config
        assert not os.path.exists(
            "/var/lib/cloud/instance/network-config.json"
        )
        initer.apply_network_config(False)
        assert f"{atomic_helper.json_dumps(netcfg)}\n" == util.load_text_file(
            "/var/lib/cloud/instance/network-config.json"
        )

    def test_none_ds_runs_modules_which_do_not_define_distros(self, caplog):
        """Any modules which do not define a distros attribute are run."""
        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )

        mods = Modules(initer)
        (which_ran, failures) = mods.run_section("cloud_init_modules")
        assert not failures
        assert os.path.exists("/etc/blah.ini")
        assert "write_files" in which_ran
        contents = util.load_text_file("/etc/blah.ini")
        assert contents == "blah"
        assert (
            "Skipping modules ['write_files'] because they are not verified on"
            " distro 'ubuntu'" not in caplog.text
        )

    def test_none_ds_skips_modules_which_define_unmatched_distros(
        self, caplog
    ):
        """Skip modules which define distros which don't match the current."""
        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )

        mods = Modules(initer)
        (which_ran, failures) = mods.run_section("cloud_init_modules")
        assert not failures
        assert (
            "Skipping modules 'spacewalk' because they are not verified on"
            " distro 'ubuntu'" in caplog.text
        )
        assert "spacewalk" not in which_ran

    def test_none_ds_runs_modules_which_distros_all(self, caplog):
        """Skip modules which define distros attribute as supporting 'all'.

        This is done in the module with the declaration:
        distros = [ALL_DISTROS]. runcmd is an example.
        """
        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )

        mods = Modules(initer)
        (which_ran, failures) = mods.run_section("cloud_init_modules")
        assert not failures
        assert "runcmd" in which_ran
        assert (
            "Skipping modules 'runcmd' because they are not verified on"
            " distro 'ubuntu'" not in caplog.text
        )

    def test_none_ds_forces_run_via_unverified_modules(self, caplog, cfg):
        """run_section forced skipped modules by using unverified_modules."""

        # re-write cloud.cfg with unverified_modules override
        cfg = copy.deepcopy(cfg)
        cfg["unverified_modules"] = ["spacewalk"]  # Would have skipped
        cloud_cfg = safeyaml.dumps(cfg)
        util.ensure_dir(os.path.join("/etc", "cloud"))
        util.write_file(os.path.join("/etc", "cloud", "cloud.cfg"), cloud_cfg)

        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )

        mods = Modules(initer)
        (which_ran, failures) = mods.run_section("cloud_init_modules")
        assert not failures
        assert "spacewalk" in which_ran
        assert "running unverified_modules: 'spacewalk'" in caplog.text

    def test_none_ds_run_with_no_config_modules(self, cfg):
        """run_section will report no modules run when none are configured."""

        # re-write cloud.cfg with unverified_modules override
        cfg = copy.deepcopy(cfg)
        # Represent empty configuration in /etc/cloud/cloud.cfg
        cfg["cloud_init_modules"] = None
        cloud_cfg = safeyaml.dumps(cfg)
        util.ensure_dir(os.path.join("/etc", "cloud"))
        util.write_file(os.path.join("/etc", "cloud", "cloud.cfg"), cloud_cfg)

        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )

        mods = Modules(initer)
        (which_ran, failures) = mods.run_section("cloud_init_modules")
        assert not failures
        assert [] == which_ran



# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit import util
from tests.unittests.helpers import get_distro


class TestArch:
    def test_get_distro(self, tmp_path):
        distro = get_distro("arch")
        hostname = "myhostname"
        hostfile = tmp_path / "hostfile"
        distro._write_hostname(hostname, hostfile)
        assert hostname + "\n" == util.load_text_file(hostfile)



# This file is part of cloud-init. See LICENSE file for license information.
import pytest

from cloudinit import atomic_helper, util
from tests.unittests.helpers import get_distro


class TestGentoo:
    def test_write_hostname(self, tmp_path):
        distro = get_distro("gentoo")
        hostname = "myhostname"
        hostfile = tmp_path / "hostfile"
        distro._write_hostname(hostname, hostfile)
        if distro.uses_systemd():
            assert "myhostname\n" == util.load_text_file(hostfile)
        else:
            assert 'hostname="myhostname"\n' == util.load_text_file(hostfile)

    def test_write_existing_hostname_with_comments(self, tmp_path):
        distro = get_distro("gentoo")
        hostname = "myhostname"
        contents = '#This is the hostname\nhostname="localhost"'
        hostfile = tmp_path / "hostfile"
        atomic_helper.write_file(hostfile, contents, omode="w")
        distro._write_hostname(hostname, hostfile)
        if distro.uses_systemd():
            assert (
                "#This is the hostname\nmyhostname\n"
                == util.load_text_file(hostfile)
            )
        else:
            assert (
                '#This is the hostname\nhostname="myhostname"\n'
                == util.load_text_file(hostfile)
            )


@pytest.fixture
def no_systemd(mocker):
    mocker.patch("cloudinit.distros.uses_systemd", return_value=False)


@pytest.mark.usefixtures("no_systemd")
class TestGentooOpenRC(TestGentoo):
    pass



# Copyright (C) 2024-2025, Raspberry Pi Ltd.
#
# Author: Paul Oberosler <paul.oberosler@raspberrypi.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging
import os
from typing import Union

from cloudinit import subp
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.settings import PER_INSTANCE

LOG = logging.getLogger(__name__)
RPI_BASE_KEY = "rpi"
RPI_INTERFACES_KEY = "interfaces"
ENABLE_USB_GADGET_KEY = "enable_usb_gadget"
SUPPORTED_INTERFACES = {
    "spi": "do_spi",
    "i2c": "do_i2c",
    "onewire": "do_onewire",
}
SERIAL_INTERFACE = "serial"
RASPI_CONFIG_SERIAL_CONS_FN = "do_serial_cons"
RASPI_CONFIG_SERIAL_HW_FN = "do_serial_hw"
RPI_USB_GADGET_SCRIPT = "/usr/bin/rpi-usb-gadget"
REBOOT_MSG = "Rebooting to apply config.txt changes..."

meta: MetaSchema = {
    "id": "cc_raspberry_pi",
    "distros": ["raspberry-pi-os"],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": [RPI_BASE_KEY],
}


def configure_usb_gadget(enable: bool) -> bool:
    """Enable / Disable Raspberry Pi USB gadget mode when
    the script is present.

    Returns True when a reboot is required.
    """
    LOG.debug("Enable rpi-usb-gadget mode: %s", enable)

    if not os.path.exists(RPI_USB_GADGET_SCRIPT):
        LOG.error("rpi-usb-gadget script not found: %s", RPI_USB_GADGET_SCRIPT)
        return False

    try:
        subp.subp(
            [
                RPI_USB_GADGET_SCRIPT,
                "on" if enable else "off",
                # Force even if the device isn't supported
                # to avoid timeout in action request
                # and to allow activation with microSD card
                # in a unsupported device because its maybe faster
                # and then the user might switch it to a
                # supported device later
                "-f",
            ],
            capture=False,
            timeout=15,
        )

        return True
    except subp.ProcessExecutionError as e:
        LOG.error("Failed to configure rpi-usb-gadget: %s", e)
        return False


def is_pifive() -> bool:
    try:
        subp.subp(["/usr/bin/raspi-config", "nonint", "is_pifive"])
        return True
    except subp.ProcessExecutionError:
        return False


def configure_serial_interface(cfg: Union[dict, bool]) -> bool:
    """Configure the serial (UART) interface on Raspberry Pi.

    Depending on the configuration, this enables or disables
    the serial console and the hardware UART by invoking
    `raspi-config` non-interactively.

    Args:
        cfg: Either a boolean or a dictionary:
            - If a bool, enables or disables the serial
              console and hardware UART together.
            - If a dict, may contain the keys `console`
              and `hardware` with boolean values controlling
              each individually.

    Returns:
        True if the configuration was applied successfully,
        False otherwise.
    """

    def get_bool_field(cfg_dict: dict, name: str, default=False):
        val = cfg_dict.get(name, default)
        if not isinstance(val, bool):
            LOG.warning(
                "Invalid value for %s.serial.%s: %s",
                RPI_INTERFACES_KEY,
                name,
                val,
            )
            return default
        return val

    enable_console = False
    enable_hw = False

    if isinstance(cfg, dict):
        enable_console = get_bool_field(cfg, "console")
        enable_hw = get_bool_field(cfg, "hardware")

    elif isinstance(cfg, bool):
        # default to enabling console as if < pi5
        # this will also enable the hardware
        enable_console = cfg

    if not is_pifive() and enable_console:
        # only pi5 has 2 usable UARTs
        # on other models, enabling the console
        # will also block the other UART
        enable_hw = True

    try:
        subp.subp(
            [
                "/usr/bin/raspi-config",
                "nonint",
                RASPI_CONFIG_SERIAL_CONS_FN,
                str(0 if enable_console else 1),
            ]
        )
    except subp.ProcessExecutionError as e:
        LOG.error("Failed to configure serial console: %s", e)
        return False

    try:
        subp.subp(
            [
                "/usr/bin/raspi-config",
                "nonint",
                RASPI_CONFIG_SERIAL_HW_FN,
                str(0 if enable_hw else 1),
            ]
        )
        return True
    except subp.ProcessExecutionError as e:
        LOG.error("Failed to configure serial hardware: %s", e)
        return False


def configure_interface(iface: str, enable: bool) -> bool:
    """Enable or disable a supported hardware interface.

    Supported interfaces are defined in `SUPPORTED_INTERFACES` and
    currently include SPI, I2C, and 1-Wire. The function applies
    the change using `raspi-config` in non-interactive mode.

    Args:
        iface: The interface name (e.g. `"spi"` or `"i2c"`).
        enable: True to enable the interface, False to disable it.

    Returns:
        True if the interface configuration succeeded, False otherwise.
    """
    assert (
        iface in SUPPORTED_INTERFACES.keys()
    ), f"Unsupported interface: {iface}"

    try:
        subp.subp(
            [
                "/usr/bin/raspi-config",
                "nonint",
                SUPPORTED_INTERFACES[iface],
                str(0 if enable else 1),
            ]
        )

        return True
    except subp.ProcessExecutionError as e:
        LOG.error("Failed to configure %s: %s", iface, e)
        return False


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    if RPI_BASE_KEY not in cfg:
        return
    elif not isinstance(cfg[RPI_BASE_KEY], dict):
        LOG.warning(
            "Invalid value for %s: %s",
            RPI_BASE_KEY,
            cfg[RPI_BASE_KEY],
        )
        return
    elif not cfg[RPI_BASE_KEY]:
        LOG.debug("Empty value for %s. Skipping...", RPI_BASE_KEY)
        return

    want_reboot = False

    for key in cfg[RPI_BASE_KEY]:
        if key == ENABLE_USB_GADGET_KEY:
            enable = cfg[RPI_BASE_KEY][key]

            if isinstance(enable, bool):
                want_reboot |= configure_usb_gadget(enable)
            else:
                raise ValueError(f"Invalid value for {ENABLE_USB_GADGET_KEY}")
            continue
        elif key == RPI_INTERFACES_KEY:
            if not isinstance(cfg[RPI_BASE_KEY][key], dict):
                LOG.warning(
                    "Invalid value for %s: %s",
                    RPI_BASE_KEY,
                    cfg[RPI_BASE_KEY][key],
                )
                return
            elif not cfg[RPI_BASE_KEY][key]:
                LOG.debug("Empty value for %s. Skipping...", key)
                return

            subkeys = list(cfg[RPI_BASE_KEY][key].keys())

            # check for supported ARM interfaces
            for subkey in subkeys:
                if (
                    subkey not in SUPPORTED_INTERFACES.keys()
                    and subkey != SERIAL_INTERFACE
                ):
                    LOG.warning(
                        "Invalid key for %s: %s", RPI_INTERFACES_KEY, subkey
                    )
                    continue

                enable = cfg[RPI_BASE_KEY][key][subkey]

                if subkey == SERIAL_INTERFACE:
                    if not isinstance(enable, (dict, bool)):
                        LOG.warning(
                            "Invalid value for %s.%s: %s",
                            RPI_INTERFACES_KEY,
                            subkey,
                            enable,
                        )
                    else:
                        want_reboot |= configure_serial_interface(enable)
                    continue

                if isinstance(enable, bool):
                    want_reboot |= configure_interface(subkey, enable)
                else:
                    LOG.warning(
                        "Invalid value for %s.%s: %s",
                        RPI_INTERFACES_KEY,
                        subkey,
                        enable,
                    )
        else:
            LOG.warning("Unsupported key: %s", key)
            continue

    if want_reboot is True:
        # Reboot to apply changes to config.txt and modprobe
        cmd = cloud.distro.shutdown_command(
            mode="reboot",
            delay="now",
            message=REBOOT_MSG,
        )
        subp.subp(cmd)



# This file is part of cloud-init. See LICENSE file for license information.

"""Tests for cloudinit.log"""

import datetime
import io
import logging
import time
from typing import cast

import pytest

from cloudinit import lifecycle, util
from cloudinit.analyze.dump import CLOUD_INIT_ASCTIME_FMT
from cloudinit.log import loggers


@pytest.fixture
def ci_logs():
    return io.StringIO()


@pytest.fixture
def log(ci_logs):
    # set up a logger like cloud-init does in setup_logging, but instead
    # of sys.stderr, we'll plug in a StringIO() object so we can see
    # what gets logged
    logging.Formatter.converter = time.gmtime
    ci_root = logging.getLogger()
    console = logging.StreamHandler(ci_logs)
    console.setFormatter(logging.Formatter(loggers.DEFAULT_LOG_FORMAT))
    console.setLevel(logging.DEBUG)
    ci_root.addHandler(console)
    ci_root.setLevel(logging.DEBUG)
    LOG = logging.getLogger("test_cloudinit_logger")
    return LOG


class TestCloudInitLogger:

    def test_logger_uses_gmtime(self, log, ci_logs):
        """Test that log message have timestamp in UTC (gmtime)"""

        # Log a message, extract the timestamp from the log entry
        # convert to datetime, and compare to a utc timestamp before
        # and after the logged message.

        # Due to loss of precision in the LOG timestamp, subtract and add
        # time to the utc stamps for comparison
        #
        # utc_before: 2017-08-23 14:19:42.569299
        # parsed dt : 2017-08-23 14:19:43.069000
        # utc_after : 2017-08-23 14:19:43.570064

        def remove_tz(_dt: datetime.datetime) -> datetime.datetime:
            """
            Removes the timezone object from an aware datetime dt without
            conversion of date and time data
            """
            return _dt.replace(tzinfo=None)

        utc_before = remove_tz(
            datetime.datetime.now(datetime.timezone.utc)
            - datetime.timedelta(0, 0.5)
        )
        log.error("Test message")
        utc_after = remove_tz(
            datetime.datetime.now(datetime.timezone.utc)
            + datetime.timedelta(0, 0.5)
        )

        # extract timestamp from log:
        # 2017-08-23 14:19:43,069 - test_log.py[ERROR]: Test message
        logstr = ci_logs.getvalue().splitlines()[0]
        timestampstr = logstr.split(" - ")[0]
        parsed_dt = datetime.datetime.strptime(
            timestampstr, CLOUD_INIT_ASCTIME_FMT
        )

        assert utc_before < parsed_dt
        assert parsed_dt < utc_after
        assert utc_before < utc_after
        assert utc_after > parsed_dt


class TestDeprecatedLogs:
    def test_deprecated_log_level(self, caplog):
        logger = cast(loggers.CustomLoggerType, logging.getLogger())
        logger.deprecated("deprecated message")
        assert "DEPRECATED" == caplog.records[0].levelname
        assert "deprecated message" in caplog.text

    def test_trace_log_level(self, caplog):
        logger = cast(loggers.CustomLoggerType, logging.getLogger())
        logger.setLevel(logging.NOTSET)
        logger.trace("trace message")
        assert "TRACE" == caplog.records[0].levelname
        assert "trace message" in caplog.text

    @pytest.mark.parametrize(
        "expected_log_level, deprecation_info_boundary",
        (
            pytest.param(
                "DEPRECATED",
                "19.2",
                id="test_same_deprecation_info_boundary_is_deprecated_level",
            ),
            pytest.param(
                "INFO",
                "19.1",
                id="test_lower_deprecation_info_boundary_is_info_level",
            ),
        ),
    )
    def test_deprecate_log_level_based_on_features(
        self,
        expected_log_level,
        deprecation_info_boundary,
        caplog,
        mocker,
        clear_deprecation_log,
    ):
        """Deprecation log level depends on key deprecation_version

        When DEPRECATION_INFO_BOUNDARY is set to a version number, and a key
        has a deprecated_version with a version greater than the boundary
        the log level is INFO instead of DEPRECATED. If
        DEPRECATION_INFO_BOUNDARY is set to the default, "devel", all
        deprecated keys are logged at level DEPRECATED.
        """
        mocker.patch.object(
            util.features,
            "DEPRECATION_INFO_BOUNDARY",
            deprecation_info_boundary,
        )
        lifecycle.deprecate(
            deprecated="some key",
            deprecated_version="19.2",
            extra_message="dont use it",
        )
        assert expected_log_level == caplog.records[0].levelname
        assert (
            "some key is deprecated in 19.2 and scheduled to be removed in"
            " 24.2" in caplog.text
        )

    def test_log_deduplication(self, caplog):
        loggers.define_extra_loggers()
        lifecycle.deprecate(
            deprecated="stuff",
            deprecated_version="19.1",
            extra_message=":)",
        )
        lifecycle.deprecate(
            deprecated="stuff",
            deprecated_version="19.1",
            extra_message=":)",
        )
        lifecycle.deprecate(
            deprecated="stuff",
            deprecated_version="19.1",
            extra_message=":)",
            schedule=6,
        )
        assert 2 == len(caplog.records)


def test_logger_prints_to_stderr(capsys):
    message = "to stdout"
    loggers.setup_basic_logging()
    logging.getLogger().warning(message)
    assert message in capsys.readouterr().err



# This file is part of cloud-init. See LICENSE file for license information.

import pytest

import cloudinit.config.cc_raspberry_pi as cc_rpi
from cloudinit.config.cc_raspberry_pi import (
    ENABLE_USB_GADGET_KEY,
    RPI_BASE_KEY,
    RPI_INTERFACES_KEY,
    RPI_USB_GADGET_SCRIPT,
)
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from cloudinit.subp import ProcessExecutionError
from tests.unittests.helpers import mock, skipUnlessJsonSchema
from tests.unittests.util import get_cloud

M_PATH = "cloudinit.config.cc_raspberry_pi."


class TestHandleRaspberryPi:
    @mock.patch(M_PATH + "configure_usb_gadget")
    def test_handle_usb_gadget_enabled(self, m_usb_gadget):
        cloud = get_cloud("raspberry_pi_os")
        cfg = {RPI_BASE_KEY: {ENABLE_USB_GADGET_KEY: True}}
        cc_rpi.handle("cc_raspberry_pi", cfg, cloud, [])
        m_usb_gadget.assert_called_once_with(True)

    @mock.patch(M_PATH + "configure_interface")
    def test_handle_configure_interface_i2c(self, m_iface):
        cloud = get_cloud("raspberry_pi_os")
        cfg = {RPI_BASE_KEY: {RPI_INTERFACES_KEY: {"i2c": True}}}
        cc_rpi.handle("cc_raspberry_pi", cfg, cloud, [])
        m_iface.assert_called_once_with("i2c", True)

    @mock.patch(M_PATH + "configure_serial_interface")
    @mock.patch(M_PATH + "is_pifive", return_value=True)
    def test_handle_configure_serial_interface_dict(self, m_ispi5, m_serial):
        cloud = get_cloud("raspberry_pi_os")
        serial_value = {
            "console": True,
            "hardware": True,
        }
        cfg = {RPI_BASE_KEY: {RPI_INTERFACES_KEY: {"serial": serial_value}}}
        cc_rpi.handle("cc_raspberry_pi", cfg, cloud, [])
        m_serial.assert_called_once_with(serial_value)

    @mock.patch(M_PATH + "configure_serial_interface")
    @mock.patch(M_PATH + "is_pifive", return_value=True)
    def test_handle_configure_serial_interface_bool(self, m_ispi5, m_serial):
        cloud = get_cloud("raspberry_pi_os")
        cfg = {RPI_BASE_KEY: {RPI_INTERFACES_KEY: {"serial": True}}}
        cc_rpi.handle("cc_raspberry_pi", cfg, cloud, [])
        m_serial.assert_called_once_with(True)

    @mock.patch(
        "cloudinit.distros.raspberry_pi_os.Distro.shutdown_command",
        return_value=["shutdown", "-r", "now", cc_rpi.REBOOT_MSG],
    )
    @mock.patch("cloudinit.subp.subp")
    @mock.patch(M_PATH + "is_pifive", return_value=True)
    def test_trigger_reboot(self, is_pi5, m_subp, m_shutdown):
        keys = list(cc_rpi.SUPPORTED_INTERFACES.keys()) + [
            cc_rpi.SERIAL_INTERFACE
        ]
        cloud = get_cloud("raspberry_pi_os")

        for key in keys:
            cfg1 = {RPI_BASE_KEY: {RPI_INTERFACES_KEY: {key: True}}}

            m_subp.reset_mock()
            m_shutdown.reset_mock()

            cc_rpi.handle("cc_raspberry_pi", cfg1, cloud, [])

            # Reboot requested via shutdown command
            m_shutdown.assert_called_once()

            # subp calls: raspi-config(s) + shutdown
            expected_calls = 3 if key == cc_rpi.SERIAL_INTERFACE else 2
            assert m_subp.call_count == expected_calls

            # Last call is the shutdown
            assert m_subp.call_args == mock.call(
                ["shutdown", "-r", "now", cc_rpi.REBOOT_MSG]
            )

        # enable_usb_gadget path: ensure script exists so the code runs
        with mock.patch(M_PATH + "os.path.exists", return_value=True):
            cfg2 = {RPI_BASE_KEY: {ENABLE_USB_GADGET_KEY: True}}

            m_subp.reset_mock()
            m_shutdown.reset_mock()

            cc_rpi.handle("cc_raspberry_pi", cfg2, cloud, [])

            m_shutdown.assert_called_once()
            # gadget script + shutdown
            assert m_subp.call_count == 2
            assert m_subp.call_args == mock.call(
                ["shutdown", "-r", "now", cc_rpi.REBOOT_MSG]
            )


class TestRaspberryPiMethods:
    @mock.patch("cloudinit.subp.subp")
    def test_configure_usb_gadget_enable(self, m_subp):
        with mock.patch("os.path.exists", return_value=True):
            cc_rpi.configure_usb_gadget(True)
        m_subp.assert_called_once_with(
            [RPI_USB_GADGET_SCRIPT, "on", "-f"], capture=False, timeout=15
        )

    @mock.patch("cloudinit.subp.subp")
    def test_configure_usb_gadget_missing_script(self, m_subp, caplog):
        """If the rpi-usb-gadget script is missing, log an error
        and return False."""
        # Simulate missing rpi-usb-gadget script
        with mock.patch("os.path.exists", return_value=False):
            with caplog.at_level("ERROR"):
                # Should not raise
                result = cc_rpi.configure_usb_gadget(True)

        # No subprocess call should be made
        m_subp.assert_not_called()

        # Verify an error was logged
        assert "rpi-usb-gadget script not found" in caplog.text

        # Reboot should not be requested
        assert result is False

    @mock.patch("cloudinit.subp.subp")
    def test_configure_usb_gadget_script_failure(self, m_subp, caplog):
        """If the rpi-usb-gadget script fails, log an error
        and return False."""
        m_subp.side_effect = cc_rpi.subp.ProcessExecutionError(
            cmd=[RPI_USB_GADGET_SCRIPT, "on", "-f"],
            exit_code=1,
            stdout="",
            stderr="fail",
        )

        with mock.patch("os.path.exists", return_value=True):
            with caplog.at_level("ERROR"):
                result = cc_rpi.configure_usb_gadget(True)

        # Subprocess should have been invoked once
        m_subp.assert_called_once_with(
            [RPI_USB_GADGET_SCRIPT, "on", "-f"], capture=False, timeout=15
        )

        # Error log should contain failure message
        assert "Failed to configure rpi-usb-gadget" in caplog.text

        # Function should return False (no reboot triggered)
        assert result is False

    @mock.patch("cloudinit.subp.subp", return_value=("ok", ""))
    def test_is_pifive_true(self, m_subp):
        assert cc_rpi.is_pifive() is True

    @mock.patch(
        "cloudinit.subp.subp",
        side_effect=ProcessExecutionError("1", [], "fail"),
    )
    def test_is_pifive_false(self, m_subp):
        assert cc_rpi.is_pifive() is False

    @mock.patch("cloudinit.subp.subp")
    def test_configure_interface_valid(self, m_subp):
        cc_rpi.configure_interface("i2c", True)
        m_subp.assert_called_once_with(
            ["/usr/bin/raspi-config", "nonint", "do_i2c", "0"]
        )

    def test_configure_interface_invalid(self):
        with pytest.raises(AssertionError):
            cc_rpi.configure_interface("invalid_iface", True)

    @mock.patch("cloudinit.subp.subp")
    @mock.patch(M_PATH + "is_pifive", return_value=True)
    def test_configure_serial_interface_dict_config(self, m_ispi5, m_subp):
        cfg = {"console": True, "hardware": False}

        # Simulate is_pifive returning True to prevent enable_hw override
        cc_rpi.configure_serial_interface(cfg)

        expected_calls = [
            mock.call(
                [
                    "/usr/bin/raspi-config",
                    "nonint",
                    cc_rpi.RASPI_CONFIG_SERIAL_CONS_FN,
                    "0",
                ]
            ),
            mock.call(
                [
                    "/usr/bin/raspi-config",
                    "nonint",
                    cc_rpi.RASPI_CONFIG_SERIAL_HW_FN,
                    "1",
                ]
            ),
        ]
        m_subp.assert_has_calls(expected_calls, any_order=False)

    @mock.patch("cloudinit.subp.subp")
    @mock.patch(M_PATH + "is_pifive", return_value=False)
    def test_configure_serial_interface_boolean_config_non_pi5(
        self, m_ispi5, m_subp
    ):
        cc_rpi.configure_serial_interface(True)

        expected_calls = [
            mock.call(
                [
                    "/usr/bin/raspi-config",
                    "nonint",
                    cc_rpi.RASPI_CONFIG_SERIAL_CONS_FN,
                    "0",
                ]
            ),
            mock.call(
                [
                    "/usr/bin/raspi-config",
                    "nonint",
                    cc_rpi.RASPI_CONFIG_SERIAL_HW_FN,
                    "0",
                ]
            ),
        ]
        m_subp.assert_has_calls(expected_calls, any_order=False)


@skipUnlessJsonSchema()
class TestRaspberryPiSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            (
                {
                    RPI_BASE_KEY: {
                        RPI_INTERFACES_KEY: {"spi": True, "i2c": False}
                    }
                },
                None,
            ),
            (
                {RPI_BASE_KEY: {RPI_INTERFACES_KEY: {"spi": "true"}}},
                f"{RPI_BASE_KEY}.{RPI_INTERFACES_KEY}.spi: 'true'"
                " is not of type 'boolean'",
            ),
            (
                {
                    RPI_BASE_KEY: {
                        RPI_INTERFACES_KEY: {
                            "serial": {"console": True, "hardware": False}
                        }
                    }
                },
                None,
            ),
            (
                {
                    RPI_BASE_KEY: {
                        RPI_INTERFACES_KEY: {"serial": {"console": 123}}
                    }
                },
                f"{RPI_BASE_KEY}.{RPI_INTERFACES_KEY}.serial.console: "
                "123 is not of type 'boolean'",
            ),
            ({RPI_BASE_KEY: {ENABLE_USB_GADGET_KEY: True}}, None),
            (
                {RPI_BASE_KEY: {ENABLE_USB_GADGET_KEY: "true"}},
                f"{RPI_BASE_KEY}.{ENABLE_USB_GADGET_KEY}: 'true'"
                " is not of type 'boolean'",
            ),
        ],
    )
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)



import unittest.mock as mock

import pytest

from tests.unittests.helpers import get_distro

try:
    # Blowfish not available in < 3.7, so this has never worked. Ignore failure
    # to import with AttributeError. We need this module imported prior to
    # patching the object, so we can't gate the version with pytest skipif.
    import cloudinit.distros.netbsd
except AttributeError:
    pass

M_PATH = "cloudinit.distros.netbsd."


class TestNetBSD:
    @mock.patch(M_PATH + "subp.subp")
    def test_add_user(self, m_subp):
        distro = get_distro("netbsd")
        assert True is distro.add_user("me2", uid=1234, default=False)
        assert [
            mock.call(
                ["useradd", "-m", "me2"], logstring=["useradd", "-m", "me2"]
            )
        ] == m_subp.call_args_list

    @mock.patch(M_PATH + "subp.subp")
    def test_unlock_passwd(self, m_subp, caplog):
        distro = get_distro("netbsd")
        distro.unlock_passwd("me2")
        assert [
            mock.call(["usermod", "-C", "no", "me2"])
        ] == m_subp.call_args_list


@pytest.mark.parametrize("with_pkgin", (True, False))
@mock.patch("cloudinit.distros.netbsd.os")
def test_init(m_os, with_pkgin):
    print(with_pkgin)
    m_os.path.exists.return_value = with_pkgin
    cfg = {}

    # patch ifconfig -a
    with mock.patch(
        "cloudinit.distros.networking.subp.subp", return_value=("", None)
    ):
        distro = cloudinit.distros.netbsd.NetBSD("netbsd", cfg, None)
    expectation = ["pkgin", "-y", "full-upgrade"] if with_pkgin else None
    assert distro.pkg_cmd_upgrade_prefix == expectation
    assert [mock.call("/usr/pkg/bin/pkgin")] == m_os.path.exists.call_args_list



# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit import util
from tests.unittests.helpers import get_distro, mock

SYSTEM_INFO = {
    "paths": {
        "cloud_dir": "/var/lib/cloud/",
        "templates_dir": "/etc/cloud/templates/",
    },
    "network": {"renderers": "networkd"},
}


class TestPhoton:
    distro = get_distro("photon", SYSTEM_INFO)
    expected_log_line = "Rely on PhotonOS default network config"

    def test_network_renderer(self):
        assert self.distro._cfg["network"]["renderers"] == "networkd"

    def test_get_distro(self):
        assert self.distro.osfamily == "photon"

    @mock.patch("cloudinit.distros.photon.subp.subp")
    def test_write_hostname(self, m_subp, caplog, tmp_path):
        hostname = "myhostname"
        hostfile = str(tmp_path / "previous-hostname")
        self.distro._write_hostname(hostname, hostfile)
        assert hostname == util.load_text_file(hostfile)

        ret = self.distro._read_hostname(hostfile)
        assert ret == hostname

        m_subp.return_value = (None, None)
        hostfile += "hostfile"
        self.distro._write_hostname(hostname, hostfile)

        m_subp.return_value = (hostname, None)
        ret = self.distro._read_hostname(hostfile)
        assert ret == hostname

        caplog.clear()
        m_subp.return_value = (None, "bla")
        self.distro._write_hostname(hostname, None)
        assert "Error while setting hostname" in caplog.text

    @mock.patch("cloudinit.net.generate_fallback_config")
    def test_fallback_netcfg(self, m_fallback_cfg, caplog):

        key = "disable_fallback_netcfg"
        # Don't use fallback if no setting given
        caplog.clear()
        assert self.distro.generate_fallback_config() is None
        assert self.expected_log_line in caplog.text

        caplog.clear()
        self.distro._cfg[key] = True
        assert self.distro.generate_fallback_config() is None
        assert self.expected_log_line in caplog.text

        caplog.clear()
        self.distro._cfg[key] = False
        assert self.distro.generate_fallback_config() is not None
        assert self.expected_log_line not in caplog.text
